import logging
from typing import Iterator, Optional

from . import expr
from .const import IRConst
from .enums import IRCallee, IRRegArray, VEXObject, get_enum_from_int, get_int_from_enum
from .errors import PyVEXError
from .expr import Const, Get, IRExpr
from .native import ffi, pvc

log = logging.getLogger("pyvex.stmt")


class IRStmt(VEXObject):
    """
    IR statements in VEX represents operations with side-effects.
    """

    tag: Optional[str] = None
    tag_int = 0  # set automatically at bottom of file

    __slots__ = []

    def pp(self):
        print(str(self))

    @property
    def child_expressions(self) -> Iterator["IRExpr"]:
        for k in self.__slots__:
            v = getattr(self, k)
            if isinstance(v, IRExpr):
                # return itself
                yield v
                # return all the child expressions
                yield from v.child_expressions

    # ???
    @property
    def expressions(self):
        return self.child_expressions

    @property
    def constants(self):
        return sum((e.constants for e in self.expressions), [])

    @staticmethod
    def _from_c(c_stmt):
        if c_stmt[0] == ffi.NULL:
            return None

        try:
            stmt_class = enum_to_stmt_class(c_stmt.tag)._from_c(c_stmt)
        except KeyError:
            raise PyVEXError("Unknown/unsupported IRStmtTag %s.\n" % get_enum_from_int(c_stmt.tag))
        return stmt_class._from_c(c_stmt)

    def typecheck(self, tyenv):  # pylint: disable=unused-argument,no-self-use
        return True

    def replace_expression(self, replacements):
        """
        Replace child expressions in-place.

        :param Dict[IRExpr, IRExpr] replacements:  A mapping from expression-to-find to expression-to-replace-with
        :return:                    None
        """

        for k in self.__slots__:
            v = getattr(self, k)
            if isinstance(v, IRExpr) and v in replacements:
                setattr(self, k, replacements.get(v))
            elif isinstance(v, IRExpr):
                v.replace_expression(replacements)
            elif type(v) is tuple:
                # Rebuild the tuple
                _lst = []
                replaced = False
                for expr_ in v:
                    if isinstance(expr_, IRExpr) and expr_ in replacements:
                        _lst.append(replacements.get(expr_))
                        replaced = True
                    else:
                        _lst.append(expr_)
                if replaced:
                    setattr(self, k, tuple(_lst))

    def __str__(self):
        return self._pp_str(None, None, None)

    def _pp_str(self, reg_name=None, arch=None, tyenv=None) -> str:
        raise NotImplementedError()


class NoOp(IRStmt):
    """
    A no-operation statement. It is usually the result of an IR optimization.
    """

    __slots__ = []

    tag = "Ist_NoOp"

    def _pp_str(self, reg_name=None, arch=None, tyenv=None):
        return "IR-NoOp"

    @staticmethod
    def _from_c(c_stmt):
        return NoOp()


class IMark(IRStmt):
    """
    An instruction mark. It marks the start of the statements that represent a single machine instruction (the end of
    those statements is marked by the next IMark or the end of the IRSB).  Contains the address and length of the
    instruction.
    """

    __slots__ = ["addr", "len", "delta"]

    tag = "Ist_IMark"

    def __init__(self, addr: int, length: int, delta: int):
        self.addr = addr
        self.len = length
        self.delta = delta

    def _pp_str(self, reg_name=None, arch=None, tyenv=None):
        return "------ IMark(0x%x, %d, %d) ------" % (self.addr, self.len, self.delta)

    @staticmethod
    def _from_c(c_stmt):
        return IMark(c_stmt.Ist.IMark.addr, c_stmt.Ist.IMark.len, c_stmt.Ist.IMark.delta)


class AbiHint(IRStmt):
    """
    An ABI hint, provides specific information about this platform's ABI.
    """

    __slots__ = ["base", "len", "nia"]

    tag = "Ist_AbiHint"

    def __init__(self, base, length, nia):
        self.base = base
        self.len = length
        self.nia = nia

    def _pp_str(self, reg_name=None, arch=None, tyenv=None):
        return "====== AbiHint(0x%s, %d, %s) ======" % (self.base, self.len, self.nia)

    @staticmethod
    def _from_c(c_stmt):
        return AbiHint(
            IRExpr._from_c(c_stmt.Ist.AbiHint.base), c_stmt.Ist.AbiHint.len, IRExpr._from_c(c_stmt.Ist.AbiHint.nia)
        )


class Put(IRStmt):
    """
    Write to a guest register, at a fixed offset in the guest state.
    """

    __slots__ = ["data", "offset"]

    tag = "Ist_Put"

    def __init__(self, data: "IRExpr", offset):
        self.data = data
        self.offset = offset

    ## TODO: Check if result_size and arch are available before looking of arch register name
    def _pp_str(self, reg_name=None, arch=None, tyenv=None):
        if arch is not None and tyenv is not None:
            reg_name = arch.translate_register_name(self.offset, self.data.result_size(tyenv) // 8)

        if reg_name is not None:
            return f"PUT({reg_name}) = {self.data}"
        else:
            return f"PUT(offset={self.offset}) = {self.data}"

    @staticmethod
    def _from_c(c_stmt):
        return Put(IRExpr._from_c(c_stmt.Ist.Put.data), c_stmt.Ist.Put.offset)

    def typecheck(self, tyenv):
        return self.data.typecheck(tyenv)


class PutI(IRStmt):
    """
    Write to a guest register, at a non-fixed offset in the guest state.
    """

    __slots__ = ["descr", "ix", "data", "bias"]

    tag = "Ist_PutI"

    def __init__(self, descr, ix, data, bias):
        self.descr = descr
        self.ix = ix
        self.data = data
        self.bias = bias

    def _pp_str(self, reg_name=None, arch=None, tyenv=None):
        return "PutI(%s)[%s,%d] = %s" % (self.descr, self.ix, self.bias, self.data)

    @staticmethod
    def _from_c(c_stmt):
        return PutI(
            IRRegArray._from_c(c_stmt.Ist.PutI.details.descr),
            IRExpr._from_c(c_stmt.Ist.PutI.details.ix),
            IRExpr._from_c(c_stmt.Ist.PutI.details.data),
            c_stmt.Ist.PutI.details.bias,
        )

    def typecheck(self, tyenv):
        dataty = self.data.typecheck(tyenv)
        if dataty is None:
            return False
        if dataty != self.descr.elemTy:
            log.debug("Expression doesn't match RegArray type")
            return False
        return True


class WrTmp(IRStmt):
    """
    Assign a value to a temporary.  Note that SSA rules require each tmp is only assigned to once.  IR sanity checking
    will reject any block containing a temporary which is not assigned to exactly once.
    """

    __slots__ = ["data", "tmp"]

    tag = "Ist_WrTmp"

    def __init__(self, tmp, data: "IRExpr"):
        self.tmp = tmp
        self.data = data

    def _pp_str(self, reg_name=None, arch=None, tyenv=None):
        # Support for named register in string representation of expr.Get

        if arch is not None and tyenv is not None and isinstance(self.data, Get):
            reg_name = arch.translate_register_name(self.data.offset, self.data.result_size(tyenv) // 8)

        if reg_name is not None and isinstance(self.data, expr.Get):
            return "t%d = %s" % (self.tmp, self.data._pp_str(reg_name=reg_name))
        else:
            return "t%d = %s" % (self.tmp, self.data)

    @staticmethod
    def _from_c(c_stmt):
        return WrTmp(c_stmt.Ist.WrTmp.tmp, IRExpr._from_c(c_stmt.Ist.WrTmp.data))

    def typecheck(self, tyenv):
        dataty = self.data.typecheck(tyenv)
        if dataty is None:
            return False
        if dataty != tyenv.lookup(self.tmp):
            log.debug("Expression doesn't match tmp type")
            return False
        return True


class Store(IRStmt):
    """
    Write a value to memory..
    """

    __slots__ = ["addr", "data", "end"]

    tag = "Ist_Store"

    def __init__(self, addr: "IRExpr", data: "IRExpr", end: str):
        self.addr = addr
        self.data = data
        self.end = end

    @property
    def endness(self):
        return self.end

    def _pp_str(self, reg_name=None, arch=None, tyenv=None):
        return f"ST{self.endness[-2:].lower()}({self.addr}) = {self.data}"

    @staticmethod
    def _from_c(c_stmt):
        return Store(
            IRExpr._from_c(c_stmt.Ist.Store.addr),
            IRExpr._from_c(c_stmt.Ist.Store.data),
            get_enum_from_int(c_stmt.Ist.Store.end),
        )

    def typecheck(self, tyenv):
        dataty = self.data.typecheck(tyenv)
        if dataty is None:
            return False
        addrty = self.addr.typecheck(tyenv)
        if addrty is None:
            return False
        if addrty != tyenv.wordty:
            log.debug("addr must be full word for arch")
            return False
        if self.end not in ("Iend_LE", "Iend_BE"):
            log.debug("invalid endness enum")
            return False
        return True


class CAS(IRStmt):
    """
    an atomic compare-and-swap operation.
    """

    __slots__ = ["addr", "dataLo", "dataHi", "expdLo", "expdHi", "oldLo", "oldHi", "end"]

    tag = "Ist_CAS"

    def __init__(self, addr, dataLo, dataHi, expdLo, expdHi, oldLo, oldHi, end):
        self.addr = addr
        self.dataLo = dataLo
        self.dataHi = dataHi
        self.expdLo = expdLo
        self.expdHi = expdHi
        self.oldLo = oldLo
        self.oldHi = oldHi
        self.end = end

    @property
    def endness(self):
        return self.end

    def _pp_str(self, reg_name=None, arch=None, tyenv=None):
        return "t({},{}) = CAS{}({} :: ({},{})->({},{}))".format(
            self.oldLo, self.oldHi, self.end[-2:].lower(), self.addr, self.expdLo, self.expdHi, self.dataLo, self.dataHi
        )

    @staticmethod
    def _from_c(c_stmt):
        return CAS(
            IRExpr._from_c(c_stmt.Ist.CAS.details.addr),
            IRExpr._from_c(c_stmt.Ist.CAS.details.dataLo),
            IRExpr._from_c(c_stmt.Ist.CAS.details.dataHi),
            IRExpr._from_c(c_stmt.Ist.CAS.details.expdLo),
            IRExpr._from_c(c_stmt.Ist.CAS.details.expdHi),
            c_stmt.Ist.CAS.details.oldLo,
            c_stmt.Ist.CAS.details.oldHi,
            get_enum_from_int(c_stmt.Ist.CAS.details.end),
        )

    def typecheck(self, tyenv):
        addrty = self.addr.typecheck(tyenv)
        if addrty is None:
            return False
        if addrty != tyenv.wordty:
            log.debug("addr must be full word for arch")
            return False
        if self.end not in ("Iend_LE", "Iend_BE"):
            log.debug("invalid endness enum")
            return False

        if self.oldHi == 0xFFFFFFFF:
            # single-element case
            if self.expdHi is not None or self.dataHi is not None:
                log.debug("expdHi and dataHi must be None")
                return False
            expdLoTy = self.expdLo.typecheck(tyenv)
            dataLoTy = self.dataLo.typecheck(tyenv)
            if expdLoTy is None or dataLoTy is None:
                return False
            if tyenv.lookup(self.oldLo) != expdLoTy or expdLoTy != dataLoTy:
                log.debug("oldLo, expdL, dataLo must all have the same type")
                return False
        else:
            # double-element case
            expdLoTy = self.expdLo.typecheck(tyenv)
            dataLoTy = self.dataLo.typecheck(tyenv)
            expdHiTy = self.expdHi.typecheck(tyenv)
            dataHiTy = self.dataHi.typecheck(tyenv)
            if expdLoTy is None or dataLoTy is None or expdHiTy is None or dataHiTy is None:
                return False
            if (
                tyenv.lookup(self.oldLo) != expdLoTy
                or expdLoTy != dataLoTy
                or tyenv.lookup(self.oldHi) != expdHiTy
                or expdHiTy != dataHiTy
                or expdLoTy != expdHiTy
            ):
                log.debug("oldLo, expdLo, dataLo, oldHi, expdHi, dataHi must all have the same type")
                return False

        return True


class LLSC(IRStmt):
    """
    Either Load-Linked or Store-Conditional, depending on STOREDATA. If STOREDATA is NULL then this is a Load-Linked,
    else it is a Store-Conditional.
    """

    __slots__ = ["addr", "storedata", "result", "end"]

    tag = "Ist_LLSC"

    def __init__(self, addr, storedata, result, end):
        self.addr = addr
        self.storedata = storedata
        self.result = result
        self.end = end

    @property
    def endness(self):
        return self.end

    def _pp_str(self, reg_name=None, arch=None, tyenv=None):
        if self.storedata is None:
            return "t%d = LD%s-Linked(%s)" % (self.result, self.end[-2:].lower(), self.addr)
        else:
            return "t%d = ( ST%s-Cond(%s) = %s )" % (self.result, self.end[-2:].lower(), self.addr, self.storedata)

    @staticmethod
    def _from_c(c_stmt):
        return LLSC(
            IRExpr._from_c(c_stmt.Ist.LLSC.addr),
            IRExpr._from_c(c_stmt.Ist.LLSC.storedata),
            c_stmt.Ist.LLSC.result,
            get_enum_from_int(c_stmt.Ist.LLSC.end),
        )

    def typecheck(self, tyenv):
        addrty = self.addr.typecheck(tyenv)
        if addrty is None:
            return False
        if addrty != tyenv.wordty:
            log.debug("addr must be full word for arch")
            return False
        if self.end not in ("Iend_LE", "Iend_BE"):
            log.debug("invalid endness enum")
            return False

        if self.storedata is not None:
            # load-linked
            storety = self.storedata.typecheck(tyenv)
            if storety is None:
                return False

            if tyenv.lookup(self.result) != "Ity_I1":
                log.debug("result tmp must be Ity_I1")
                return False

        return True


class MBE(IRStmt):
    __slots__ = ["event"]

    tag = "Ist_MBE"

    def __init__(self, event):
        self.event = event

    def _pp_str(self, reg_name=None, arch=None, tyenv=None):
        return "MBusEvent-" + self.event

    @staticmethod
    def _from_c(c_stmt):
        return MBE(get_enum_from_int(c_stmt.Ist.MBE.event))


class Dirty(IRStmt):
    __slots__ = ["cee", "guard", "args", "tmp", "mFx", "mAddr", "mSize", "nFxState"]

    tag = "Ist_Dirty"

    def __init__(self, cee, guard, args, tmp, mFx, mAddr, mSize, nFxState):
        self.cee = cee
        self.guard = guard
        self.args = tuple(args)
        self.tmp = tmp
        self.mFx = mFx
        self.mAddr = mAddr
        self.mSize = mSize
        self.nFxState = nFxState

    def _pp_str(self, reg_name=None, arch=None, tyenv=None):
        return "t{} = DIRTY {} {} ::: {}({})".format(
            self.tmp, self.guard, "TODO(effects)", self.cee, ",".join(str(a) for a in self.args)
        )

    @property
    def child_expressions(self):
        expressions = sum((a.child_expressions for a in self.args), [])
        expressions.extend(self.args)
        expressions.append(self.guard)
        expressions.extend(self.guard.child_expressions)
        return expressions

    @staticmethod
    def _from_c(c_stmt):
        args = []
        for i in range(20):
            a = c_stmt.Ist.Dirty.details.args[i]
            if a == ffi.NULL:
                break

            args.append(IRExpr._from_c(a))

        return Dirty(
            IRCallee._from_c(c_stmt.Ist.Dirty.details.cee),
            IRExpr._from_c(c_stmt.Ist.Dirty.details.guard),
            tuple(args),
            c_stmt.Ist.Dirty.details.tmp,
            get_enum_from_int(c_stmt.Ist.Dirty.details.mFx),
            IRExpr._from_c(c_stmt.Ist.Dirty.details.mAddr),
            c_stmt.Ist.Dirty.details.mSize,
            c_stmt.Ist.Dirty.details.nFxState,
        )


class Exit(IRStmt):
    """
    A conditional exit from the middle of an IRSB.
    """

    __slots__ = ["guard", "dst", "offsIP", "jk"]

    tag = "Ist_Exit"

    def __init__(self, guard, dst, jk, offsIP):
        self.guard = guard
        self.dst = dst
        self.offsIP = offsIP
        self.jk = jk

    @property
    def jumpkind(self):
        return self.jk

    def _pp_str(self, reg_name=None, arch=None, tyenv=None):
        if arch is not None and tyenv is not None:
            reg_name = arch.translate_register_name(self.offsIP, arch.bits // 8)

        if reg_name is None:
            return "if (%s) { PUT(offset=%d) = %#x; %s }" % (self.guard, self.offsIP, self.dst.value, self.jumpkind)
        else:
            return f"if ({self.guard}) {{ PUT({reg_name}) = {self.dst.value:#x}; {self.jumpkind} }}"

    @property
    def child_expressions(self):
        return [self.guard] + self.guard.child_expressions + [Const(self.dst)]

    @staticmethod
    def _from_c(c_stmt):
        return Exit(
            IRExpr._from_c(c_stmt.Ist.Exit.guard),
            IRConst._from_c(c_stmt.Ist.Exit.dst),
            get_enum_from_int(c_stmt.Ist.Exit.jk),
            c_stmt.Ist.Exit.offsIP,
        )

    def typecheck(self, tyenv):
        if not self.jk.startswith("Ijk_"):
            log.debug("Jumpkind is not a jumpkind enum")
            return False
        guardty = self.guard.typecheck(tyenv)
        if guardty is None:
            return False
        if guardty != "Ity_I1":
            log.debug("guard must be Ity_I1")
            return False
        return True


class LoadG(IRStmt):
    """
    A guarded load.
    """

    __slots__ = ["addr", "alt", "guard", "dst", "cvt", "end", "cvt_types"]

    tag = "Ist_LoadG"

    def __init__(self, end, cvt, dst, addr, alt, guard):
        self.addr = addr
        self.alt = alt
        self.guard = guard
        self.dst = dst
        self.cvt = cvt
        self.end = end

        type_in = ffi.new("IRType *")  # TODO separate this from the pyvex C implementation
        type_out = ffi.new("IRType *")
        pvc.typeOfIRLoadGOp(get_int_from_enum(self.cvt), type_out, type_in)
        type_in = ffi.cast("int *", type_in)[0]
        type_out = ffi.cast("int *", type_out)[0]
        self.cvt_types = (get_enum_from_int(type_in), get_enum_from_int(type_out))

    @property
    def endness(self):
        return self.end

    def _pp_str(self, reg_name=None, arch=None, tyenv=None):
        return "t%d = if (%s) %s(LD%s(%s)) else %s" % (
            self.dst,
            self.guard,
            self.cvt,
            self.end[-2:].lower(),
            self.addr,
            self.alt,
        )

    @staticmethod
    def _from_c(c_stmt):
        return LoadG(
            get_enum_from_int(c_stmt.Ist.LoadG.details.end),
            get_enum_from_int(c_stmt.Ist.LoadG.details.cvt),
            c_stmt.Ist.LoadG.details.dst,
            IRExpr._from_c(c_stmt.Ist.LoadG.details.addr),
            IRExpr._from_c(c_stmt.Ist.LoadG.details.alt),
            IRExpr._from_c(c_stmt.Ist.LoadG.details.guard),
        )

    def typecheck(self, tyenv):
        addrty = self.addr.typecheck(tyenv)
        if addrty is None:
            return False
        if addrty != tyenv.wordty:
            log.debug("addr must be full word for arch")
            return False
        if self.end not in ("Iend_LE", "Iend_BE"):
            log.debug("invalid endness enum")
            return False

        dstty = tyenv.lookup(self.dst)
        guardty = self.guard.typecheck(tyenv)
        altty = self.alt.typecheck(tyenv)

        if guardty is None or altty is None:
            return False
        if dstty != "Ity_I32" or altty != "Ity_I32":
            log.debug("dst and alt must be Ity_I32")
            return False
        if guardty != "Ity_I1":
            log.debug("guard must be Ity_I1")
            return False
        if not self.cvt.startswith("ILGop_"):
            log.debug("Invalid cvt enum")
            return False
        return True


class StoreG(IRStmt):
    """
    A guarded store.
    """

    __slots__ = ["addr", "data", "guard", "end"]

    tag = "Ist_StoreG"

    def __init__(self, end, addr, data, guard):
        self.addr = addr
        self.data = data
        self.guard = guard
        self.end = end

    @property
    def endness(self):
        return self.end

    def _pp_str(self, reg_name=None, arch=None, tyenv=None):
        return f"if ({self.guard}) ST{self.end[-2:].lower()}({self.addr}) = {self.data}"

    @staticmethod
    def _from_c(c_stmt):
        return StoreG(
            get_enum_from_int(c_stmt.Ist.StoreG.details.end),
            IRExpr._from_c(c_stmt.Ist.StoreG.details.addr),
            IRExpr._from_c(c_stmt.Ist.StoreG.details.data),
            IRExpr._from_c(c_stmt.Ist.StoreG.details.guard),
        )

    def typecheck(self, tyenv):
        addrty = self.addr.typecheck(tyenv)
        if addrty is None:
            return False
        if addrty != tyenv.wordty:
            log.debug("addr must be full word for arch")
            return False
        if self.end not in ("Iend_LE", "Iend_BE"):
            log.debug("invalid endness enum")
            return False

        guardty = self.guard.typecheck(tyenv)
        dataty = self.data.typecheck(tyenv)

        if guardty is None or dataty is None:
            return False
        if guardty != "Ity_I1":
            log.debug("guard must be Ity_I1")
            return False
        return True


_globals = globals().copy()
#
# Mapping from tag strings/enums to IRStmt classes
#
tag_to_stmt_mapping = {}
enum_to_stmt_mapping = {}
tag_count = 0
cls = None
for cls in _globals.values():
    if type(cls) is type and issubclass(cls, IRStmt) and cls is not IRStmt:
        tag_to_stmt_mapping[cls.tag] = cls
        enum_to_stmt_mapping[get_int_from_enum(cls.tag)] = cls
        cls.tag_int = tag_count
        tag_count += 1
del cls


def tag_to_stmt_class(tag):
    try:
        return tag_to_stmt_mapping[tag]
    except KeyError:
        raise KeyError("No statement class for tag %s." % tag)


def enum_to_stmt_class(tag_enum):
    try:
        return enum_to_stmt_mapping[tag_enum]
    except KeyError:
        raise KeyError("No statement class for tag %s." % get_enum_from_int(tag_enum))



import copy
import re

from pyvex.const import U1, get_type_size, ty_to_const_class, vex_int_class
from pyvex.enums import IRCallee
from pyvex.expr import ITE, Binop, CCall, Const, Get, Load, RdTmp, Unop
from pyvex.stmt import Dirty, Exit, IMark, NoOp, Put, Store, WrTmp


class JumpKind:
    Boring = "Ijk_Boring"
    Call = "Ijk_Call"
    Ret = "Ijk_Ret"
    Segfault = "Ijk_SigSEGV"
    Exit = "Ijk_Exit"
    Syscall = "Ijk_Sys_syscall"
    Sysenter = "Ijk_Sys_sysenter"
    Invalid = "Ijk_INVALID"
    NoDecode = "Ijk_NoDecode"


class TypeMeta(type):
    typemeta_re = re.compile(r"int_(?P<size>\d+)$")

    def __getattr__(self, name):
        match = self.typemeta_re.match(name)
        if match:
            width = int(match.group("size"))
            return vex_int_class(width).type
        else:
            return type.__getattr__(name)


class Type(metaclass=TypeMeta):
    __metaclass__ = TypeMeta

    ieee_float_16 = "Ity_F16"
    ieee_float_32 = "Ity_F32"
    ieee_float_64 = "Ity_F64"
    ieee_float_128 = "Ity_F128"
    decimal_float_32 = "Ity_D32"
    decimal_float_64 = "Ity_D64"
    decimal_float_128 = "Ity_D128"
    simd_vector_128 = "Ity_V128"
    simd_vector_256 = "Ity_V256"


def get_op_format_from_const_ty(ty):
    return ty_to_const_class(ty).op_format


def make_format_op_generator(fmt_string):
    """
    Return a function which generates an op format (just a string of the vex instruction)

    Functions by formatting the fmt_string with the types of the arguments
    """

    def gen(arg_types):
        converted_arg_types = list(map(get_op_format_from_const_ty, arg_types))
        op = fmt_string.format(arg_t=converted_arg_types)
        return op

    return gen


def mkbinop(fstring):
    return lambda self, expr_a, expr_b: self.op_binary(make_format_op_generator(fstring))(expr_a, expr_b)


def mkunop(fstring):
    return lambda self, expr_a: self.op_unary(make_format_op_generator(fstring))(expr_a)


def mkcmpop(fstring_fragment, signedness=""):
    def cmpop(self, expr_a, expr_b):
        ty = self.get_type(expr_a)
        fstring = f"Iop_Cmp{fstring_fragment}{{arg_t[0]}}{signedness}"
        retval = mkbinop(fstring)(self, expr_a, expr_b)
        return self.cast_to(retval, ty)

    return cmpop


class IRSBCustomizer:
    op_add = mkbinop("Iop_Add{arg_t[0]}")
    op_sub = mkbinop("Iop_Sub{arg_t[0]}")
    op_umul = mkbinop("Iop_Mul{arg_t[0]}")
    op_smul = mkbinop("Iop_MullS{arg_t[0]}")
    op_sdiv = mkbinop("Iop_DivS{arg_t[0]}")
    op_udiv = mkbinop("Iop_DivU{arg_t[0]}")

    # Custom operation that does not exist in libVEX
    op_mod = mkbinop("Iop_Mod{arg_t[0]}")

    op_or = mkbinop("Iop_Or{arg_t[0]}")
    op_and = mkbinop("Iop_And{arg_t[0]}")
    op_xor = mkbinop("Iop_Xor{arg_t[0]}")

    op_shr = mkbinop("Iop_Shr{arg_t[0]}")  # Shift Right (logical)
    op_shl = mkbinop("Iop_Shl{arg_t[0]}")  # Shift Left (logical)

    op_sar = mkbinop("Iop_Sar{arg_t[0]}")  # Shift Arithmetic Right operation

    op_not = mkunop("Iop_Not{arg_t[0]}")

    op_cmp_eq = mkcmpop("EQ")
    op_cmp_ne = mkcmpop("NE")
    op_cmp_slt = mkcmpop("LT", "S")
    op_cmp_sle = mkcmpop("LE", "S")
    op_cmp_ult = mkcmpop("LT", "U")
    op_cmp_ule = mkcmpop("LE", "U")
    op_cmp_sge = mkcmpop("GE", "S")
    op_cmp_uge = mkcmpop("GE", "U")
    op_cmp_sgt = mkcmpop("GT", "S")
    op_cmp_ugt = mkcmpop("GT", "U")

    def __init__(self, irsb):
        self.arch = irsb.arch
        self.irsb = irsb

    def get_type(self, rdt):
        return rdt.result_type(self.irsb.tyenv)

    # Statements (no return value)
    def _append_stmt(self, stmt):
        self.irsb.statements += [stmt]

    def imark(self, int_addr, int_length, int_delta=0):
        self._append_stmt(IMark(int_addr, int_length, int_delta))

    def get_reg(self, regname):  # TODO move this into the lifter
        return self.arch.registers[regname][0]

    def put(self, expr_val, tuple_reg):
        self._append_stmt(Put(copy.copy(expr_val), tuple_reg))

    def store(self, addr, expr):
        self._append_stmt(Store(copy.copy(addr), copy.copy(expr), self.arch.memory_endness))

    def noop(self):
        self._append_stmt(NoOp())

    def add_exit(self, guard, dst, jk, ip):
        """
        Add an exit out of the middle of an IRSB.
        (e.g., a conditional jump)
        :param guard: An expression, the exit is taken if true
        :param dst: the destination of the exit (a Const)
        :param jk: the JumpKind of this exit (probably Ijk_Boring)
        :param ip: The address of this exit's source
        """
        self.irsb.statements.append(Exit(guard, dst.con, jk, ip))

    # end statements

    def goto(self, addr):
        self.irsb.next = addr
        self.irsb.jumpkind = JumpKind.Boring

    def ret(self, addr):
        self.irsb.next = addr
        self.irsb.jumpkind = JumpKind.Ret

    def call(self, addr):
        self.irsb.next = addr
        self.irsb.jumpkind = JumpKind.Call

    def _add_tmp(self, t):
        return self.irsb.tyenv.add(t)

    def _rdtmp(self, tmp):
        return RdTmp.get_instance(tmp)

    def _settmp(self, expr):
        ty = self.get_type(expr)
        tmp = self._add_tmp(ty)
        self._append_stmt(WrTmp(tmp, expr))
        return self._rdtmp(tmp)

    def rdreg(self, reg, ty):
        return self._settmp(Get(reg, ty))

    def load(self, addr, ty):
        return self._settmp(Load(self.arch.memory_endness, ty, copy.copy(addr)))

    def op_ccall(self, retty, funcstr, args):
        return self._settmp(CCall(retty, IRCallee(len(args), funcstr, 0xFFFF), args))

    def dirty(self, retty, funcstr, args):
        if retty is None:
            tmp = 0xFFFFFFFF
        else:
            tmp = self._add_tmp(retty)
        self._append_stmt(Dirty(IRCallee(len(args), funcstr, 0xFFFF), Const(U1(1)), args, tmp, None, None, None, None))
        return self._rdtmp(tmp)

    def ite(self, condrdt, iftruerdt, iffalserdt):
        return self._settmp(ITE(copy.copy(condrdt), copy.copy(iffalserdt), copy.copy(iftruerdt)))

    def mkconst(self, val, ty):
        cls = ty_to_const_class(ty)
        return Const(cls(val))

    # Operations
    def op_generic(self, Operation, op_generator):
        def instance(*args):  # Note: The args here are all RdTmps
            for arg in args:
                assert isinstance(arg, RdTmp) or isinstance(arg, Const)
            arg_types = [self.get_type(arg) for arg in args]
            # two operations should never share the same argument instances, copy them here to ensure that
            args = [copy.copy(a) for a in args]
            op = Operation(op_generator(arg_types), args)
            msg = "operation needs to be well typed: " + str(op)
            assert op.typecheck(self.irsb.tyenv), msg + "\ntypes: " + str(self.irsb.tyenv)
            return self._settmp(op)

        return instance

    def op_binary(self, op_format_str):
        return self.op_generic(Binop, op_format_str)

    def op_unary(self, op_format_str):
        return self.op_generic(Unop, op_format_str)

    def cast_to(self, rdt, tydest, signed=False, high=False):
        goalwidth = get_type_size(tydest)
        rdtwidth = self.get_rdt_width(rdt)

        if rdtwidth > goalwidth:
            return self.op_narrow_int(rdt, tydest, high_half=high)
        elif rdtwidth < goalwidth:
            return self.op_widen_int(rdt, tydest, signed=signed)
        else:
            return rdt

    def op_to_one_bit(self, rdt):
        rdtty = self.get_type(rdt)
        if rdtty not in [Type.int_64, Type.int_32]:
            rdt = self.op_widen_int_unsigned(rdt, Type.int_32)
        onebit = self.op_narrow_int(rdt, Type.int_1)
        return onebit

    def op_narrow_int(self, rdt, tydest, high_half=False):
        op_name = "{op}{high}to{dest}".format(
            op="Iop_{arg_t[0]}", high="HI" if high_half else "", dest=get_op_format_from_const_ty(tydest)
        )
        return self.op_unary(make_format_op_generator(op_name))(rdt)

    def op_widen_int(self, rdt, tydest, signed=False):
        op_name = "{op}{sign}to{dest}".format(
            op="Iop_{arg_t[0]}", sign="S" if signed else "U", dest=get_op_format_from_const_ty(tydest)
        )
        return self.op_unary(make_format_op_generator(op_name))(rdt)

    def op_widen_int_signed(self, rdt, tydest):
        return self.op_widen_int(rdt, tydest, signed=True)

    def op_widen_int_unsigned(self, rdt, tydest):
        return self.op_widen_int(rdt, tydest, signed=False)

    def get_msb(self, tmp, ty):
        width = get_type_size(ty)
        return self.get_bit(tmp, width - 1)

    def get_bit(self, rdt, idx):
        shifted = self.op_shr(rdt, idx)
        bit = self.op_extract_lsb(shifted)
        return bit

    def op_extract_lsb(self, rdt):
        bitmask = self.mkconst(1, self.get_type(rdt))
        return self.op_and(bitmask, rdt)

    def set_bit(self, rdt, idx, bval):
        currbit = self.get_bit(rdt, idx)
        areequalextrabits = self.op_xor(bval, currbit)
        one = self.mkconst(1, self.get_type(areequalextrabits))
        areequal = self.op_and(areequalextrabits, one)
        shifted = self.op_shl(areequal, idx)
        return self.op_xor(rdt, shifted)

    def set_bits(self, rdt, idxsandvals):
        ty = self.get_type(rdt)
        if all([isinstance(idx, Const) for idx, _ in idxsandvals]):
            relevantbits = self.mkconst(sum([1 << idx.con.value for idx, _ in idxsandvals]), ty)
        else:
            relevantbits = self.mkconst(0, ty)
            for idx, _ in idxsandvals:
                shifted = self.op_shl(self.mkconst(1, ty), idx)
                relevantbits = self.op_or(relevantbits, shifted)
        setto = self.mkconst(0, ty)
        for idx, bval in idxsandvals:
            bvalbit = self.op_extract_lsb(bval)
            shifted = self.op_shl(bvalbit, idx)
            setto = self.op_or(setto, shifted)
        shouldflip = self.op_and(self.op_xor(setto, rdt), relevantbits)
        return self.op_xor(rdt, shouldflip)

    def get_rdt_width(self, rdt):
        return rdt.result_size(self.irsb.tyenv)



import logging
import re
from typing import Dict, List, Optional, Tuple

from .const import U8, U16, U32, U64, IRConst, get_type_size
from .enums import IRCallee, IRRegArray, VEXObject, get_enum_from_int, get_int_from_enum
from .errors import PyVEXError
from .native import ffi, pvc

log = logging.getLogger("pyvex.expr")


class IRExpr(VEXObject):
    """
    IR expressions in VEX represent operations without side effects.
    """

    __slots__ = []

    tag: Optional[str] = None
    tag_int = 0  # set automatically at bottom of file

    def pp(self):
        print(str(self))

    def __str__(self):
        return self._pp_str()

    def _pp_str(self) -> str:
        raise NotImplementedError

    @property
    def child_expressions(self) -> List["IRExpr"]:
        """
        A list of all of the expressions that this expression ends up evaluating.
        """
        expressions = []
        for k in self.__slots__:
            v = getattr(self, k)
            if isinstance(v, IRExpr):
                expressions.append(v)
                expressions.extend(v.child_expressions)
        return expressions

    @property
    def constants(self):
        """
        A list of all of the constants that this expression ends up using.
        """
        constants = []
        for k in self.__slots__:
            v = getattr(self, k)
            if isinstance(v, IRExpr):
                constants.extend(v.constants)
            elif isinstance(v, IRConst):
                constants.append(v)
        return constants

    def result_size(self, tyenv):
        return get_type_size(self.result_type(tyenv))

    def result_type(self, tyenv):
        raise NotImplementedError()

    def replace_expression(self, replacements):
        """
        Replace child expressions in-place.

        :param Dict[IRExpr, IRExpr] replacements:  A mapping from expression-to-find to expression-to-replace-with
        :return:                    None
        """

        for k in self.__slots__:
            v = getattr(self, k)
            if isinstance(v, IRExpr) and v in replacements:
                setattr(self, k, replacements.get(v))
            elif isinstance(v, list):
                # Replace the instance in the list
                for i, expr_ in enumerate(v):
                    if isinstance(expr_, IRExpr) and expr_ in replacements:
                        v[i] = replacements.get(expr_)
            elif type(v) is tuple:
                # Rebuild the tuple
                _lst = []
                replaced = False
                for i, expr_ in enumerate(v):
                    if isinstance(expr_, IRExpr) and expr_ in replacements:
                        _lst.append(replacements.get(expr_))
                        replaced = True
                    else:
                        _lst.append(expr_)
                if replaced:
                    setattr(self, k, tuple(_lst))
            elif isinstance(v, IRExpr):
                v.replace_expression(replacements)

    @staticmethod
    def _from_c(c_expr) -> Optional["IRExpr"]:
        if c_expr == ffi.NULL or c_expr[0] == ffi.NULL:
            return None

        try:
            return enum_to_expr_class(c_expr.tag)._from_c(c_expr)
        except KeyError:
            raise PyVEXError("Unknown/unsupported IRExprTag %s\n" % get_enum_from_int(c_expr.tag))

    _translate = _from_c

    @staticmethod
    def _to_c(expr):
        try:
            return tag_to_expr_class(expr.tag)._to_c(expr)
        except KeyError:
            raise PyVEXError("Unknown/unsupported IRExprTag %s\n" % expr.tag)

    def typecheck(self, tyenv):
        return self.result_type(tyenv)


class Binder(IRExpr):
    """
    Used only in pattern matching within Vex. Should not be seen outside of Vex.
    """

    __slots__ = ["binder"]

    tag = "Iex_Binder"

    def __init__(self, binder):
        self.binder = binder

    def _pp_str(self):
        return "Binder"

    @staticmethod
    def _from_c(c_expr):
        return Binder(c_expr.iex.Binder.binder)

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_Binder(expr.binder)

    def result_type(self, tyenv):
        return "Ity_INVALID"


class VECRET(IRExpr):
    tag = "Iex_VECRET"

    __slots__ = []

    def _pp_str(self):
        return "VECRET"

    @staticmethod
    def _from_c(c_expr):
        return VECRET()

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_VECRET()

    def result_type(self, tyenv):
        return "Ity_INVALID"


class GSPTR(IRExpr):
    __slots__ = []

    tag = "Iex_GSPTR"

    def _pp_str(self):
        return "GSPTR"

    @staticmethod
    def _from_c(c_expr):
        return GSPTR()

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_GSPTR()

    def result_type(self, tyenv):
        return "Ity_INVALID"


class GetI(IRExpr):
    """
    Read a guest register at a non-fixed offset in the guest state.
    """

    __slots__ = ["descr", "ix", "bias"]

    tag = "Iex_GetI"

    def __init__(self, descr, ix, bias):
        self.descr = descr
        self.ix = ix
        self.bias = bias

    @property
    def description(self):
        return self.descr

    @property
    def index(self):
        return self.ix

    def _pp_str(self):
        return f"GetI({self.descr})[{self.ix},{self.bias}]"

    @staticmethod
    def _from_c(c_expr):
        descr = IRRegArray._from_c(c_expr.Iex.GetI.descr)
        ix = IRExpr._from_c(c_expr.Iex.GetI.ix)
        bias = c_expr.Iex.GetI.bias
        return GetI(descr, ix, bias)

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_GetI(IRRegArray._to_c(expr.descr), IRExpr._to_c(expr.ix), expr.bias)

    def result_type(self, tyenv):
        return self.descr.elemTy


class RdTmp(IRExpr):
    """
    Read the value held by a temporary.
    """

    __slots__ = ["_tmp"]

    tag = "Iex_RdTmp"

    def __init__(self, tmp):
        self._tmp = tmp

    def _pp_str(self):
        return "t%d" % self.tmp

    @property
    def tmp(self):
        return self._tmp

    @staticmethod
    def _from_c(c_expr):
        tmp = c_expr.Iex.RdTmp.tmp
        return RdTmp.get_instance(tmp)

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_RdTmp(expr.tmp)

    @staticmethod
    def get_instance(tmp):
        if tmp < 1024:
            # for small tmp reads, they are cached and are only created once globally
            return _RDTMP_POOL[tmp]
        return RdTmp(tmp)

    def replace_expression(self, replacements):
        # RdTmp is one of the terminal IRExprs, which cannot be replaced.
        pass

    def result_type(self, tyenv):
        return tyenv.lookup(self.tmp)

    def __hash__(self):
        return 133700 + self._tmp


_RDTMP_POOL = list(RdTmp(i) for i in range(0, 1024))


class Get(IRExpr):
    """
    Read a guest register, at a fixed offset in the guest state.
    """

    __slots__ = ["offset", "ty_int"]

    tag = "Iex_Get"

    def __init__(self, offset, ty: str, ty_int: Optional[int] = None):
        self.offset = offset
        if ty_int is None:
            self.ty_int = get_int_from_enum(ty)
        else:
            self.ty_int = ty_int

    @property
    def ty(self):
        return get_enum_from_int(self.ty_int)

    @property
    def type(self):
        return get_enum_from_int(self.ty_int)

    def _pp_str(self, reg_name=None):
        if reg_name:
            return f"GET:{self.ty[4:]}({reg_name})"
        else:
            return f"GET:{self.ty[4:]}(offset={self.offset})"

    @staticmethod
    def _from_c(c_expr):
        return Get(c_expr.Iex.Get.offset, get_enum_from_int(c_expr.Iex.Get.ty))

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_Get(expr.offset, expr.ty_int)

    def result_type(self, tyenv):
        return self.ty

    def __hash__(self):
        return (self.offset << 8) | self.ty_int


class Qop(IRExpr):
    """
    A quaternary operation (4 arguments).
    """

    __slots__ = ["op", "args"]

    tag = "Iex_Qop"

    def __init__(self, op, args):
        self.op = op
        self.args = args

    def _pp_str(self):
        return "{}({})".format(self.op[4:], ",".join(str(a) for a in self.args))

    @property
    def child_expressions(self):
        expressions = sum((a.child_expressions for a in self.args), [])
        expressions.extend(self.args)
        return expressions

    @staticmethod
    def _from_c(c_expr):
        return Qop(
            get_enum_from_int(c_expr.Iex.Qop.details.op),
            [
                IRExpr._from_c(arg)
                for arg in [
                    c_expr.Iex.Qop.details.arg1,
                    c_expr.Iex.Qop.details.arg2,
                    c_expr.Iex.Qop.details.arg3,
                    c_expr.Iex.Qop.details.arg4,
                ]
            ],
        )

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_Qop(get_int_from_enum(expr.op), *[IRExpr._to_c(arg) for arg in expr.args])

    def result_type(self, tyenv):
        return get_op_retty(self.op)

    def typecheck(self, tyenv):  # TODO change all this to use PyvexTypeErrorException
        resty, (arg1ty, arg2ty, arg3ty, arg4ty) = op_arg_types(self.op)
        arg1ty_real = self.args[0].typecheck(tyenv)
        arg2ty_real = self.args[1].typecheck(tyenv)
        arg3ty_real = self.args[2].typecheck(tyenv)
        arg4ty_real = self.args[3].typecheck(tyenv)
        if arg1ty_real is None or arg2ty_real is None or arg3ty_real is None or arg4ty_real is None:
            return None

        if arg1ty_real != arg1ty:
            log.debug("First arg of %s must be %s", self.op, arg1ty)
            return None
        if arg2ty_real != arg2ty:
            log.debug("Second arg of %s must be %s", self.op, arg2ty)
            return None
        if arg3ty_real != arg3ty:
            log.debug("Third arg of %s must be %s", self.op, arg3ty)
            return None
        if arg4ty_real != arg4ty:
            log.debug("Fourth arg of %s must be %s", self.op, arg4ty)
            return None

        return resty


class Triop(IRExpr):
    """
    A ternary operation (3 arguments)
    """

    __slots__ = ["op", "args"]

    tag = "Iex_Triop"

    def __init__(self, op, args):
        self.op = op
        self.args = args

    def _pp_str(self):
        return "{}({})".format(self.op[4:], ",".join(str(a) for a in self.args))

    @property
    def child_expressions(self):
        expressions = sum((a.child_expressions for a in self.args), [])
        expressions.extend(self.args)
        return expressions

    @staticmethod
    def _from_c(c_expr):
        return Triop(
            get_enum_from_int(c_expr.Iex.Triop.details.op),
            [
                IRExpr._from_c(arg)
                for arg in [c_expr.Iex.Triop.details.arg1, c_expr.Iex.Triop.details.arg2, c_expr.Iex.Triop.details.arg3]
            ],
        )

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_Triop(get_int_from_enum(expr.op), *[IRExpr._to_c(arg) for arg in expr.args])

    def result_type(self, tyenv):
        return get_op_retty(self.op)

    def typecheck(self, tyenv):
        resty, (arg1ty, arg2ty, arg3ty) = op_arg_types(self.op)
        arg1ty_real = self.args[0].typecheck(tyenv)
        arg2ty_real = self.args[1].typecheck(tyenv)
        arg3ty_real = self.args[2].typecheck(tyenv)
        if arg1ty_real is None or arg2ty_real is None or arg3ty_real is None:
            return None

        if arg1ty_real != arg1ty:
            log.debug("First arg of %s must be %s", self.op, arg1ty)
            return None
        if arg2ty_real != arg2ty:
            log.debug("Second arg of %s must be %s", self.op, arg2ty)
            return None
        if arg3ty_real != arg3ty:
            log.debug("Third arg of %s must be %s", self.op, arg3ty)
            return None

        return resty


class Binop(IRExpr):
    """
    A binary operation (2 arguments).
    """

    __slots__ = ["_op", "op_int", "args"]

    tag = "Iex_Binop"

    def __init__(self, op, args, op_int=None):
        self.op_int = op_int
        self.args = args
        self._op = op if op is not None else None

    def _pp_str(self):
        return "{}({})".format(self.op[4:], ",".join(str(a) for a in self.args))

    @property
    def op(self):
        if self._op is None:
            self._op = get_enum_from_int(self.op_int)
        return self._op

    @property
    def child_expressions(self):
        expressions = sum((a.child_expressions for a in self.args), [])
        expressions.extend(self.args)
        return expressions

    @staticmethod
    def _from_c(c_expr):
        return Binop(
            None,
            [IRExpr._from_c(arg) for arg in [c_expr.Iex.Binop.arg1, c_expr.Iex.Binop.arg2]],
            op_int=c_expr.Iex.Binop.op,
        )

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_Binop(get_int_from_enum(expr.op), *[IRExpr._to_c(arg) for arg in expr.args])

    def result_type(self, tyenv):
        return get_op_retty(self.op)

    def typecheck(self, tyenv):
        arg1ty_real = self.args[0].typecheck(tyenv)
        arg2ty_real = self.args[1].typecheck(tyenv)

        resty, (arg1ty, arg2ty) = op_arg_types(self.op)
        if arg1ty_real is None or arg2ty_real is None:
            return None

        if arg1ty_real != arg1ty:
            log.debug("First arg of %s must be %s", self.op, arg1ty)
            return None
        if arg2ty_real != arg2ty:
            log.debug("Second arg of %s must be %s", self.op, arg2ty)
            return None

        return resty


class Unop(IRExpr):
    """
    A unary operation (1 argument).
    """

    __slots__ = ["op", "args"]

    tag = "Iex_Unop"

    def __init__(self, op, args):
        self.op = op
        self.args = args

    def _pp_str(self):
        return "{}({})".format(self.op[4:], ",".join(str(a) for a in self.args))

    @property
    def child_expressions(self):
        expressions = sum((a.child_expressions for a in self.args), [])
        expressions.extend(self.args)
        return expressions

    @staticmethod
    def _from_c(c_expr):
        return Unop(get_enum_from_int(c_expr.Iex.Unop.op), [IRExpr._from_c(c_expr.Iex.Unop.arg)])

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_Unop(get_int_from_enum(expr.op), IRExpr._to_c(expr.args[0]))

    def result_type(self, tyenv):
        return get_op_retty(self.op)

    def typecheck(self, tyenv):
        resty, (arg1ty,) = op_arg_types(self.op)
        arg1ty_real = self.args[0].typecheck(tyenv)
        if arg1ty_real is None:
            return None

        if arg1ty_real != arg1ty:
            log.debug("First arg of %s must be %s", self.op, arg1ty)
            return None

        return resty


class Load(IRExpr):
    """
    A load from memory.
    """

    __slots__ = ["end", "ty", "addr"]

    tag = "Iex_Load"

    def __init__(self, end, ty, addr):
        self.end = end
        self.ty = ty
        self.addr = addr

    @property
    def endness(self):
        return self.end

    @property
    def type(self):
        return self.ty

    def _pp_str(self):
        return f"LD{self.end[-2:].lower()}:{self.ty[4:]}({self.addr})"

    @staticmethod
    def _from_c(c_expr):
        return Load(
            get_enum_from_int(c_expr.Iex.Load.end),
            get_enum_from_int(c_expr.Iex.Load.ty),
            IRExpr._from_c(c_expr.Iex.Load.addr),
        )

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_Load(get_int_from_enum(expr.end), get_int_from_enum(expr.ty), IRExpr._to_c(expr.addr))

    def result_type(self, tyenv):
        return self.ty

    def typecheck(self, tyenv):
        addrty = self.addr.typecheck(tyenv)
        if addrty is None:
            return None
        if addrty != tyenv.wordty:
            log.debug("Address must be word-sized")
            return None
        return self.ty


class Const(IRExpr):
    """
    A constant expression.
    """

    __slots__ = ["_con"]

    tag = "Iex_Const"

    def __init__(self, con: "IRConst"):
        self._con = con

    def _pp_str(self):
        return str(self.con)

    @property
    def con(self) -> "IRConst":
        return self._con

    @staticmethod
    def _from_c(c_expr):
        con = IRConst._from_c(c_expr.Iex.Const.con)
        return Const.get_instance(con)

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_Const(IRConst._to_c(expr.con))

    @staticmethod
    def get_instance(con):
        if con.value < 1024 and con.__class__ in _CONST_POOL:
            return _CONST_POOL[con.__class__][con.value]
        return Const(con)

    def result_type(self, tyenv):
        return self.con.type


_CONST_POOL = {
    U8: [Const(U8(i)) for i in range(0, 1024)],
    U16: [Const(U16(i)) for i in range(0, 1024)],
    U32: [Const(U32(i)) for i in range(0, 1024)],
    U64: [Const(U64(i)) for i in range(0, 1024)],
}


class ITE(IRExpr):
    """
    An if-then-else expression.
    """

    __slots__ = ["cond", "iffalse", "iftrue"]

    tag = "Iex_ITE"

    def __init__(self, cond, iffalse, iftrue):
        self.cond = cond
        self.iffalse = iffalse
        self.iftrue = iftrue

    def _pp_str(self):
        return f"ITE({self.cond},{self.iftrue},{self.iffalse})"

    @staticmethod
    def _from_c(c_expr):
        return ITE(
            IRExpr._from_c(c_expr.Iex.ITE.cond),
            IRExpr._from_c(c_expr.Iex.ITE.iffalse),
            IRExpr._from_c(c_expr.Iex.ITE.iftrue),
        )

    @staticmethod
    def _to_c(expr):
        return pvc.IRExpr_ITE(IRExpr._to_c(expr.cond), IRExpr._to_c(expr.iftrue), IRExpr._to_c(expr.iffalse))

    def result_type(self, tyenv):
        return self.iftrue.result_type(tyenv)

    def typecheck(self, tyenv):
        condty = self.cond.typecheck(tyenv)
        falsety = self.iffalse.typecheck(tyenv)
        truety = self.iftrue.typecheck(tyenv)

        if condty is None or falsety is None or truety is None:
            return None

        if condty != "Ity_I1":
            log.debug("guard must be Ity_I1")
            return None

        if falsety != truety:
            log.debug("false condition must be same type as true condition")
            return None

        return falsety


class CCall(IRExpr):
    """
    A call to a pure (no side-effects) helper C function.
    """

    __slots__ = ["retty", "cee", "args"]

    tag = "Iex_CCall"

    def __init__(self, retty, cee, args):
        self.retty = retty
        self.cee = cee
        self.args = tuple(args)

    @property
    def ret_type(self):
        return self.retty

    @property
    def callee(self):
        return self.cee

    def _pp_str(self):
        return "{}({}):{}".format(self.cee, ",".join(str(a) for a in self.args), self.retty)

    @property
    def child_expressions(self):
        expressions = sum((a.child_expressions for a in self.args), [])
        expressions.extend(self.args)
        return expressions

    @staticmethod
    def _from_c(c_expr):
        i = 0
        args = []
        while True:
            arg = c_expr.Iex.CCall.args[i]
            if arg == ffi.NULL:
                break
            args.append(IRExpr._from_c(arg))
            i += 1

        return CCall(get_enum_from_int(c_expr.Iex.CCall.retty), IRCallee._from_c(c_expr.Iex.CCall.cee), tuple(args))

    @staticmethod
    def _to_c(expr):
        args = [IRExpr._to_c(arg) for arg in expr.args]
        mkIRExprVec = getattr(pvc, "mkIRExprVec_%d" % len(args))
        return pvc.IRExpr_CCall(IRCallee._to_c(expr.cee), get_int_from_enum(expr.retty), mkIRExprVec(*args))

    def result_type(self, tyenv):
        return self.retty


def get_op_retty(op):
    return op_arg_types(op)[0]


op_signatures: Dict[str, Tuple[str, Tuple[str, ...]]] = {}


def _request_op_type_from_cache(op):
    return op_signatures[op]


def _request_op_type_from_libvex(op):
    Ity_INVALID = 0x1100  # as defined in enum IRType in VEX

    res_ty = ffi.new("IRType *")
    arg_tys = [ffi.new("IRType *") for _ in range(4)]
    # initialize all IRTypes to Ity_INVALID
    for arg in arg_tys:
        arg[0] = Ity_INVALID
    pvc.typeOfPrimop(get_int_from_enum(op), res_ty, *arg_tys)
    arg_ty_vals = [a[0] for a in arg_tys]

    try:
        numargs = arg_ty_vals.index(Ity_INVALID)
    except ValueError:
        numargs = 4
    args_tys_list = [get_enum_from_int(arg_ty_vals[i]) for i in range(numargs)]

    op_ty_sig = (get_enum_from_int(res_ty[0]), tuple(args_tys_list))
    op_signatures[op] = op_ty_sig
    return op_ty_sig


class PyvexOpMatchException(Exception):
    pass


class PyvexTypeErrorException(Exception):
    pass


def int_type_for_size(size):
    return "Ity_I%d" % size


# precompiled regexes
unop_signature_re = re.compile(r"Iop_(Not|Ctz|Clz)(?P<size>\d+)$")
binop_signature_re = re.compile(r"Iop_(Add|Sub|Mul|Xor|Or|And|Div[SU]|Mod)(?P<size>\d+)$")
shift_signature_re = re.compile(r"Iop_(Shl|Shr|Sar)(?P<size>\d+)$")
cmp_signature_re_1 = re.compile(r"Iop_Cmp(EQ|NE)(?P<size>\d+)$")
cmp_signature_re_2 = re.compile(r"Iop_Cmp(GT|GE|LT|LE)(?P<size>\d+)[SU]$")
mull_signature_re = re.compile(r"Iop_Mull[SU](?P<size>\d+)$")
half_signature_re = re.compile(r"Iop_DivMod[SU](?P<fullsize>\d+)to(?P<halfsize>\d+)$")
cast_signature_re = re.compile(r"Iop_(?P<srcsize>\d+)(U|S|HI|HL)?to(?P<dstsize>\d+)")


def unop_signature(op):
    m = unop_signature_re.match(op)
    if m is None:
        raise PyvexOpMatchException()
    size = int(m.group("size"))
    size_type = int_type_for_size(size)
    return size_type, (size_type,)


def binop_signature(op):
    m = binop_signature_re.match(op)
    if m is None:
        raise PyvexOpMatchException()
    size = int(m.group("size"))
    size_type = int_type_for_size(size)
    return (size_type, (size_type, size_type))


def shift_signature(op):
    m = shift_signature_re.match(op)
    if m is None:
        raise PyvexOpMatchException()
    size = int(m.group("size"))
    if size > 255:
        raise PyvexTypeErrorException("Cannot apply shift operation to %d size int because shift index is 8-bit" % size)
    size_type = int_type_for_size(size)
    return (size_type, (size_type, int_type_for_size(8)))


def cmp_signature(op):
    m = cmp_signature_re_1.match(op)
    m2 = cmp_signature_re_2.match(op)
    if (m is None) == (m2 is None):
        raise PyvexOpMatchException()
    mfound = m if m is not None else m2
    size = int(mfound.group("size"))
    size_type = int_type_for_size(size)
    return (int_type_for_size(1), (size_type, size_type))


def mull_signature(op):
    m = mull_signature_re.match(op)
    if m is None:
        raise PyvexOpMatchException()
    size = int(m.group("size"))
    size_type = int_type_for_size(size)
    doubled_size_type = int_type_for_size(2 * size)
    return (doubled_size_type, (size_type, size_type))


def half_signature(op):
    m = half_signature_re.match(op)
    if m is None:
        raise PyvexOpMatchException()
    fullsize = int(m.group("fullsize"))
    halfsize = int(m.group("halfsize"))
    if halfsize * 2 != fullsize:
        raise PyvexTypeErrorException("Invalid Instruction %s: Type 1 must be twice the size of type 2" % op)
    fullsize_type = int_type_for_size(fullsize)
    halfsize_type = int_type_for_size(halfsize)
    return (fullsize_type, (fullsize_type, halfsize_type))


def cast_signature(op):
    m = cast_signature_re.match(op)
    if m is None:
        raise PyvexOpMatchException()
    src_type = int_type_for_size(int(m.group("srcsize")))
    dst_type = int_type_for_size(int(m.group("dstsize")))
    return (dst_type, (src_type,))


polymorphic_op_processors = [
    unop_signature,
    binop_signature,
    shift_signature,
    cmp_signature,
    mull_signature,
    half_signature,
    cast_signature,
]


def _request_polymorphic_op_type(op):
    for polymorphic_signature in polymorphic_op_processors:
        try:
            op_ty_sig = polymorphic_signature(op)
            break
        except PyvexOpMatchException:
            continue
    else:
        raise PyvexOpMatchException("Op %s not recognized" % op)
    return op_ty_sig


_request_funcs = [_request_op_type_from_cache, _request_op_type_from_libvex, _request_polymorphic_op_type]


def op_arg_types(op):
    for _request_func in _request_funcs:
        try:
            return _request_func(op)
        except KeyError:
            continue
    raise ValueError("Cannot find type of op %s" % op)


_globals = globals().copy()
#
# Mapping from tag strings/enums to IRExpr classes
#
tag_to_expr_mapping = {}
enum_to_expr_mapping = {}
tag_count = 0
cls = None
for cls in _globals.values():
    if type(cls) is type and issubclass(cls, IRExpr) and cls is not IRExpr:
        tag_to_expr_mapping[cls.tag] = cls
        enum_to_expr_mapping[get_int_from_enum(cls.tag)] = cls
        cls.tag_int = tag_count
        tag_count += 1
del cls


def tag_to_expr_class(tag):
    """
    Convert a tag string to the corresponding IRExpr class type.

    :param str tag: The tag string.
    :return:        A class.
    :rtype:         type
    """

    try:
        return tag_to_expr_mapping[tag]
    except KeyError:
        raise KeyError("Cannot find expression class for type %s." % tag)


def enum_to_expr_class(tag_enum):
    """
    Convert a tag enum to the corresponding IRExpr class.

    :param int tag_enum: The tag enum.
    :return:             A class.
    :rtype:              type
    """

    try:
        return enum_to_expr_mapping[tag_enum]
    except KeyError:
        raise KeyError("Cannot find expression class for type %s." % get_enum_from_int(tag_enum))



import logging

from pyvex.lifting.util import GymratLifter, Instruction, JumpKind, Type

log = logging.getLogger(__name__)

# pylint: disable=missing-class-docstring


class Instruction_SWAPGS(Instruction):
    name = "SWAPGS"
    bin_format = "000011110000000111111000"  # 0f 01 f8

    def compute_result(self, *args):
        pass  # TODO check for priv mode


class Instruction_SYSRET(Instruction):
    name = "SYSRET"
    bin_format = "010010000000111100000111"  # 48 04 07

    def compute_result(self, *args):
        result = self.dirty(Type.int_64, "%sg_dirtyhelper_SYSRET" % self.arch.name.lower(), ())
        self.jump(None, result, JumpKind.Ret)


class Instruction_IRETQ(Instruction):
    name = "IRETQ"
    bin_format = "0100100011001111"  # 48 cf

    def compute_result(self, *args):
        result = self.dirty(Type.int_64, "%sg_dirtyhelper_IRETQ" % self.arch.name.lower(), ())
        self.jump(None, result, JumpKind.Ret)


class Instruction_RDMSR(Instruction):
    name = "RDMSR"
    bin_format = "0000111100110010"  # 0f 32

    def compute_result(self, *args):
        ecx = self.get("ecx", Type.int_32)
        result = self.dirty(Type.int_64, "%sg_dirtyhelper_RDMSR" % self.arch.name.lower(), (ecx,))
        edx = result.narrow_high(Type.int_32)
        eax = result.narrow_low(Type.int_32)
        if self.arch.bits == 32:
            self.put(eax, "eax")
            self.put(edx, "edx")
        else:
            self.put(eax.widen_unsigned(Type.int_64), "rax")
            self.put(edx.widen_unsigned(Type.int_64), "rdx")


class Instruction_XGETBV(Instruction):
    name = "XGETBV"
    bin_format = "000011110000000111010000"  # 0f 01 d0

    def compute_result(self, *args):
        ecx = self.get("ecx", Type.int_32)
        result = self.dirty(Type.int_64, "%sg_dirtyhelper_XGETBV" % self.arch.name.lower(), (ecx,))
        edx = result.narrow_high(Type.int_32)
        eax = result.narrow_low(Type.int_32)
        if self.arch.bits == 32:
            self.put(eax, "eax")
            self.put(edx, "edx")
        else:
            self.put(eax.widen_unsigned(Type.int_64), "rax")
            self.put(edx.widen_unsigned(Type.int_64), "rdx")


class Instruction_AAM(Instruction):
    name = "AAM"
    bin_format = "11010100iiiiiiii"

    # From https://www.felixcloutier.com/x86/aam
    def compute_result(self):  # pylint: disable=arguments-differ
        base = self.constant(int(self.data["i"], 2), Type.int_8)
        temp_al = self.get("al", Type.int_8)
        temp_ah = temp_al // base
        temp_al = temp_al % base
        self.put(temp_ah, "ah")
        self.put(temp_al, "al")
        log.debug(
            "The generalized AAM instruction is not supported by VEX, and is handled specially by pyvex."
            " It has no flag handling at present.  See pyvex/lifting/gym/x86_spotter.py for details"
        )

    # TODO: Flags


class Instruction_AAD(Instruction):
    name = "AAD"
    bin_format = "11010101iiiiiiii"

    # From https://www.felixcloutier.com/x86/aad
    def compute_result(self):  # pylint: disable=arguments-differ
        base = self.constant(int(self.data["i"], 2), Type.int_8)
        temp_al = self.get("al", Type.int_8)
        temp_ah = self.get("ah", Type.int_8)
        temp_al = (temp_al + (temp_ah * base)) & 0xFF
        temp_ah = self.constant(0, Type.int_8)
        self.put(temp_ah, "ah")
        self.put(temp_al, "al")
        log.debug(
            "The generalized AAD instruction is not supported by VEX, and is handled specially by pyvex."
            " It has no flag handling at present.  See pyvex/lifting/gym/x86_spotter.py for details"
        )

    # TODO: Flags


class AMD64Spotter(GymratLifter):
    instrs = [
        Instruction_RDMSR,
        Instruction_XGETBV,
        Instruction_AAD,
        Instruction_AAM,
        Instruction_SWAPGS,
        Instruction_IRETQ,
        Instruction_SYSRET,
    ]


class X86Spotter(GymratLifter):
    instrs = [
        Instruction_RDMSR,
        Instruction_XGETBV,
        Instruction_AAD,
        Instruction_AAM,
    ]
